﻿using System;
using System.Collections.Generic;
using System.Linq;

class Interpreter
{
    readonly List<Rule> AllRules;
	readonly List<Node> OrderOfExecution;
	readonly Dictionary<RuleDescription, List<Rule>> DescriptionContainTheseRules;
	readonly Rule ThePlayersCommand;
	readonly AbstractSyntaxTree Yes = new AbstractSyntaxTree("true", ValueTypes.boolean);
	readonly AbstractSyntaxTree No = new AbstractSyntaxTree("false", ValueTypes.boolean);
	int RulesDebuggingCommand = 0;
	int RulebooksDebuggingCommand = 0;

	public Interpreter(List<Node> orderOfExecution, Dictionary<RuleDescription, List<Rule>> descriptionContainTheseRules, List<Rule> allRules)
	{
		OrderOfExecution = orderOfExecution;
		DescriptionContainTheseRules = descriptionContainTheseRules;
		AllRules = allRules;
		ThePlayersCommand = AllRules.Find(r => r.Atom == "ThePlayersCommand");
	}

	public void RunProgram()
    {
		Console.WriteLine("\n=== Start Program ==========\n");
		EvalCode(AllRules.Find(r => r.Atom == "ProgramStart").Code);
		var abeyedRules = new List<List<Rule>>();
		while (string.Compare(ThePlayersCommand.Code.Value, "quit", StringComparison.InvariantCultureIgnoreCase) != 0)
	    {
			abeyedRules.Clear();
		    foreach (Node groupBeginOrEnd in OrderOfExecution)
		    {
				RuleDescription description = groupBeginOrEnd.term.RuleDescription;
				if (groupBeginOrEnd.term.Begins == 1) // BEGIN
			    {
					if (RulebooksDebuggingCommand > 0) Console.WriteLine("[Rulebook \"" + description + "\" begins.]");
					List<Rule> scheduledRules = DescriptionContainTheseRules[description];
					abeyedRules.Add(scheduledRules);
				}
			    else // END -- only run rules in this group which haven't been run already
				{
					List<Rule> previouslyScheduledRules = abeyedRules[abeyedRules.Count - 1];
					abeyedRules.RemoveAt(abeyedRules.Count - 1);
					// if group A wholly contains group B, then after group B runs its rules, do NOT re-run the B rules despite them also being in A
					foreach (Rule rule in previouslyScheduledRules)
						foreach (List<Rule> lineitem in abeyedRules)
							lineitem.Remove(rule);
					RunRulebook(previouslyScheduledRules);
					if (RulebooksDebuggingCommand > 0) Console.WriteLine("[Rulebook \"" + description + "\" ends.]");
				}
		    }
	    }
	    Console.WriteLine("\n=== End Program ===========\n");
    }

	public void RunRulebook(List<Rule> rulebook)
	{
		foreach (Rule rule in rulebook)
			if (rule.Condition == null || EvalCode(rule.Condition).Value == "true")
			{
				if (RulesDebuggingCommand > 0) Console.WriteLine("[Rule " + rule.Atom + " applies.]");
				if (rule.Code != null)
					EvalCode(rule.Code);
			}
			else if (RulesDebuggingCommand >= 2)
				Console.WriteLine("[Rule " + rule.Atom + " does NOT apply.]");
	}

	AbstractSyntaxTree EvalCode(AbstractSyntaxTree node)
    {
        if (node == null) return null;
        if (node.IsOperand) return node;
		AbstractSyntaxTree left = EvalCode(node.Left);
		if (node.Value == "?")
		{
			left = TestRelation(left);
			if (left != Yes) return null;
	    }
        AbstractSyntaxTree right = EvalCode(node.Right);
	    switch (node.Value)
	    {
			case "?":
				InvokeFunction(right);
				return null;

			case "{":
			case ";":
			    InvokeFunction(left);
			    InvokeFunction(right);
			    return null;

		    case ":=": 
			    var rule = AllRules.FirstOrDefault(r => r.Atom == left.ToString());
			    if (rule == null)
			    {
				    rule = new Rule {Atom = left.ToString(), Code = right};
				    AllRules.Add(rule);
			    }
			    if (right.Value == "ReadKeyboard")
					rule.Code = new AbstractSyntaxTree(Console.ReadLine(), ValueTypes.text);
				else if (right.Value == "ReadKey")
					rule.Code = new AbstractSyntaxTree(Console.ReadKey().KeyChar.ToString(), ValueTypes.text);
			    else
				    rule.Code = right;
				return null;

			case "(":
			    return SubstituteParamsWithValues(node);

	    }
		// at this point, any items of the shape   foo(bar,blah)   are relations, not function-invocations. We test for existance and value.
	    left = TestRelation(left);
	    right = TestRelation(right);
		// these produce a value. Statements (above) do not produce values.
		switch (node.Value)
        {
			case "<":  return left.Value.CompareTo(right.Value) < 0 ? Yes : No;
			case "<=": return left.Value.CompareTo(right.Value) <= 0 ? Yes : No;
            case "<>":
			case "!=": return left.Value.CompareTo(right.Value) != 0 ? Yes : No;
			case ">":  return left.Value.CompareTo(right.Value) > 0 ? Yes : No;
			case ">=": return left.Value.CompareTo(right.Value) >= 0 ? Yes : No;
			case "==": return left.Value.CompareTo(right.Value) == 0 ? Yes : No;
            case "&":
			case "&&": return AsBoolean(left) && AsBoolean(right) ? Yes : No;
            case "|":
			case "||": return AsBoolean(left) || AsBoolean(right) ? Yes : No;
			case "+":  return new AbstractSyntaxTree((decimal.Parse(left.Value) + decimal.Parse(right.Value)).ToString(), ValueTypes.number);
			case "-":  return new AbstractSyntaxTree((decimal.Parse(left.Value) - decimal.Parse(right.Value)).ToString(), ValueTypes.number);
			case "*":  return new AbstractSyntaxTree((decimal.Parse(left.Value) * decimal.Parse(right.Value)).ToString(), ValueTypes.number);
			case "/":  return new AbstractSyntaxTree((decimal.Parse(left.Value) / decimal.Parse(right.Value)).ToString(), ValueTypes.number);
		}
        return node;
    }

	private bool AsBoolean(AbstractSyntaxTree node)
	{
		bool retval = false;
		if (bool.TryParse(node.Value, out retval))
			return retval;
		if (node.ValueType == ValueTypes.number) return true; // a non-void rule was found
		if (node.ValueType == ValueTypes.text) return true;
		return false;
	}

	// when node.Value == "("
	AbstractSyntaxTree SubstituteParamsWithValues(AbstractSyntaxTree node)
	{
		if (node.Right == null || node.Right.Value == null) return node; // there are no parameters
		var retval = new AbstractSyntaxTree(node);
		bool hasUnbound = false;
		for (var subnode = retval.Right; subnode != null; subnode = subnode.Right)
		{
			if (subnode.Value == ",")
			{
				if (subnode.Left.Value == "_")
				{
					if (hasUnbound) throw new Exception("Can't have two _ placeholders in " + node);
					hasUnbound = true;
				}
				else
				{
					var parameter = AllRules.FirstOrDefault(r => r.Atom == subnode.Left.Value);
					if (parameter != null && parameter.Code != null) subnode.Left.Value = EvalCode(parameter.Code).ToString();
				}
			}
			else
			{
				if (subnode.Value == "_")
				{
					if (hasUnbound) throw new Exception("Can't have two _ placeholders in " + node);
					hasUnbound = true;
				}
				else
				{
					var parameter = AllRules.FirstOrDefault(r => r.Atom == subnode.Value);
					if (parameter != null && parameter.Code != null) subnode.Value = EvalCode(parameter.Code).ToString();
				}
			}
		}
		if (hasUnbound)
		{
			string relationToFind = retval.ToString();// now has the params replaced by their args
			var pieces = relationToFind.Split(new[] { '_' }, StringSplitOptions.RemoveEmptyEntries);
			Rule rule = AllRules.Find(r => r.Atom.StartsWith(pieces[0]) && r.Atom.EndsWith(pieces[1]));
			string foundValue = (rule == null) ? "" : rule.Atom.Replace(pieces[0], "").Replace(pieces[1], "");
			return new AbstractSyntaxTree(foundValue);
			//for (var subnode = retval.Right; subnode != null; subnode = subnode.Right)
			//	if (subnode.Value == "_")
			//		subnode.Value = foundValue;
		}
		return retval;
	}

	// all parameters should be bound to arguments, but the relation-name needn't actually exist
	AbstractSyntaxTree TestRelation(AbstractSyntaxTree node)
	{
		if (node.ValueType == ValueTypes.code)
		{
			string relationToFind = node.ToString();
			Rule rule = null;
			if (relationToFind.Contains("_") && relationToFind != "_")
			{
				var pieces = relationToFind.Split(new[] {'_'}, StringSplitOptions.RemoveEmptyEntries);
				if (pieces.Length > 2) throw new Exception("Can't have two _ placeholders in " + relationToFind);
				rule = AllRules.Find(r => r.Atom.StartsWith(pieces[0]) && r.Atom.EndsWith(pieces[1]));
				//if (rule != null) return new AbstractSyntaxTree(rule.Atom.Replace(pieces[0],"").Replace(pieces[1],""), ValueTypes.code);
				node = (rule != null) ? rule.Code ?? Yes : No;
			}
			else
			{
				rule = AllRules.Find(r => r.Atom == relationToFind);
				node = (rule != null) ? rule.Code ?? Yes : No;
			}
		}
		return node;
	}

	// all parameters should be bound to arguments
	void InvokeFunction(AbstractSyntaxTree node)
    {
		if (node == null || node.Value == null)
			return;
		if (RulesDebuggingCommand > 0 && node.ValueType == ValueTypes.code)
			Console.WriteLine("[Rule " + node + " invoked.]");
		if (node.Value == "Parse")
	    {
		    string input = ThePlayersCommand.Code.Value;
		    var words = input.Split(new[] { ' ' }, StringSplitOptions.RemoveEmptyEntries);
		    var verb = AllRules.Find(r => r.Atom == "verb");
		    var noun = AllRules.Find(r => r.Atom == "noun");
		    var prep = AllRules.Find(r => r.Atom == "prep");
		    var noun2 = AllRules.Find(r => r.Atom == "noun2");
			if (verb.Code == null) verb.Code = new AbstractSyntaxTree("",ValueTypes.text);
			if (noun.Code == null) noun.Code = new AbstractSyntaxTree("", ValueTypes.text);
			if (prep.Code == null) prep.Code = new AbstractSyntaxTree("", ValueTypes.text);
			if (noun2.Code == null) noun2.Code = new AbstractSyntaxTree("", ValueTypes.text);
			verb.Code.Value = (words.Length >= 1) ? words[0] : "";
			noun.Code.Value = (words.Length >= 2) ? words[1] : "";
			prep.Code.Value = (words.Length >= 3) ? words[2] : "";
			noun2.Code.Value = (words.Length >= 4) ? words[3] : "";
		    Rule referredObject;
			referredObject = AllRules.Find(r => r.Atom.StartsWith("understand(\"" + noun.Code.Value + "\""));
			if (referredObject != null) noun.Code = new AbstractSyntaxTree(referredObject.Atom.Substring(referredObject.Atom.IndexOf(',') + 1).Trim(')',' '), ValueTypes.code); 
			referredObject = AllRules.Find(r => r.Atom.StartsWith("understand(\"" + noun2.Code.Value + "\""));
			if (referredObject != null) noun2.Code = new AbstractSyntaxTree(referredObject.Atom.Substring(referredObject.Atom.IndexOf(',') + 1).Trim(')', ' '), ValueTypes.code); 

			// built-in debugging commands
		    if (words.Length >= 1 && string.Compare(words[0], "rules", StringComparison.InvariantCultureIgnoreCase) == 0)
		    {
				if (words.Length < 2)
				    RulesDebuggingCommand = RulesDebuggingCommand == 0 ? 1 : 0;
				else
				{
					if (words[1] == "on")
						RulesDebuggingCommand = 1;
					else if (words[1] == "off")
						RulesDebuggingCommand = 0;
					else if (words[1] == "all")
						RulesDebuggingCommand = 2;
				}
			    Console.WriteLine("[The RULES debugging command is now " + RulesDebuggingCommand + "]");
		    }
			if (words.Length >= 1 && string.Compare(words[0], "rulebooks", StringComparison.InvariantCultureIgnoreCase) == 0)
			{
				RulebooksDebuggingCommand = RulebooksDebuggingCommand == 0 ? 1 : 0;
				Console.WriteLine("[The RULEBOOKS debugging command is now " + RulebooksDebuggingCommand + "]");
			}
			if (words.Length >= 1 && string.Compare(words[0], "relations", StringComparison.InvariantCultureIgnoreCase) == 0)
			{
				foreach (Rule rule in AllRules)
					if (rule.Code != null && rule.Code.IsOperand)
						Console.WriteLine(rule.Atom + ": " + rule.Code);
			}
		    return;
	    }
		RuleApplies(node);
    }

	private void RuleApplies(AbstractSyntaxTree node)
	{
		if (node.ValueType != ValueTypes.code) // then run all rules with that name
			Do(node);
		else
		{
			if (node.Value != "(")
			{
				foreach (Rule invoked in AllRules.Where(r => r.Atom == node.Value))
				{
					if (invoked != null && invoked.Code != null)
						Do(invoked.Code);
				}
			}
			else
			{
				var invokedRule = AllRules.Find(r => r.Atom == node.ToString());
				if (invokedRule != null)
					Do(invokedRule.Code);
				else
					foreach (Rule invoked in AllRules.Where(r => r.Atom.StartsWith(node.Left.Value + "(")))
					{
						if (invoked != null && invoked.Code != null)
							Do(invoked.Code);
					}
			}
		}
	}

	private void Do(AbstractSyntaxTree node)
	{
		// ...and then depending on the value's type, print or execute it.
		if (node.ValueType == ValueTypes.code)
			EvalCode(node);
		else if (node.ValueType == ValueTypes.text)
			Console.Write(node.Value);
		else if (node.ValueType == ValueTypes.number)
			Console.Write(node.Value);
		else if (node.ValueType == ValueTypes.boolean)
			Console.Write(node.Value == "true" ? "yes" : "no");
		else
			throw new Exception("We never defined:  do " + node.Value + "... ");
	}
}
    